/********************************** CF92ISR.C *********************************

Copyright AT&T Global Information Solutions NCR Microelectronics Division


Author:		Erich Otto AT&T GIS System Applications Engineer
Purpose:	53CF92 Demonstration Software

This software was written by NCR Microelectronics to develop and test new
products.  It is not a general SCSI driver.  NCR assumes no liability for
its use, misuse or inability to use.

Project files:
		CF92.C 		Main body of code
		TARGET92.C	Target Mode routines
		CF92ISR.C	Interrupt Service Routine
		CF92TEST.C	Automatic Test Module
		AT_DMA92.C	Hardware dependant DMA code (8237 compatible)
		AT_DMA92.H  DMA controller parameters
		CF92AT.H	Hardware addresses, global defines
		CF92EV.H	External variables


				  **************/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "c80s_at.h"
#include "c80s_ev.h"


/********************** CF92 Interrupt Service Routine ***********************
isr(int phase)

Input argument = expected phase     0 = Selection complete, next Phase not set
				    1 = Reselection complete, next Phase not set
				    2 = DMA Data transfer complete


The interrupt service routine polls for an interrupt; reads the current SCSI
bus status, bus & status, and current SCSI data registers; decides whether
these values are the expected ones; prints "completed" on the CRT if yes or
"ERROR" and a register dump if not.  The words "completed" or ERROR are printed
on the same line as the Phase description printed by one of the subroutines that
program the 53C80S. The phase description is printed before 53C80s is enabled,
and Completed or Error is printed after the interrupt is serviced.

							   ***************/

void isr(int phase) {

	while (!kbhit() && ((inportb(stat_reg) & 0x10) == 0));	/* Poll for Interrupt or keyboard hit */

	scsi_stat = inportb(scsi_bus);
	status = inportb(stat_reg);
	cur_data  = inportb(scsi_data);
	inportb(rst_int);			/* Clear Interrupt */

	if (kbhit())
		{temp = getch();		/* flush keyboard char */
		printf("\n\nERROR!\n\nKeyboard hit while waiting for interrupt: isr(%d)",phase);
		if ((status & 0x30) == 0x30)
			printf("\n Parity Error");
		printf("\n00  Current SCSI Data		=   %02X",cur_data);
		printf("\n01  Initiator Command         =   %02X",inportb(init_cmd));
		printf("\n02  Mode                      =   %02X",inportb(mod_reg));
		printf("\n03  Target Command            =   %02X",inportb(tgt_cmd));
		printf("\n04  Current SCSI Bus Status   =   %02X",scsi_stat);
		printf("\n05  Bus & Status              =   %02X",status);
		printf("\n06  Input Data                =   %02X",inportb(in_data));
		inportb(rst_int);			/* Reset Interrupt & Parity*/

		printf("\nCDB =");
		for (i = 0; i < 16; i++)
			printf (" %02X",cdb[i]);
		scsi_error = true;
		pause_flag = true;
		return;
		}

	if (scsi_stat & 0x80)				/* SCSI Reset Detected */
		{printf("\n SCSI Reset Detected before Reselect");
		while (inportb(scsi_bus) & 0x80);
		scsi_error = true;
		return;
		}

	switch (phase) {
		case 0:										/* Selection Complete, Int, */
			if ((scsi_stat & 0x02) == 0x02 && (cur_data & parms.mid) == parms.mid)	/* SEL/ asserted & my ID (may not */
				{if(parms.detail == 'y')					/* be able to catch BSY deasserted */
					printf("complete");
				}
			else
				{printf(" ERROR !  isr(0)");
				if ((status & 0x30) == 0x30)
					printf("\n Parity Error");
				printf("\n00  Current SCSI Data   	=   %02X",cur_data);
				printf("\n01  Initiator Command         =   %02X",inportb(init_cmd));
				printf("\n02  Mode                      =   %02X",inportb(mod_reg));
				printf("\n03  Target Command            =   %02X",inportb(tgt_cmd));
				printf("\n04  Current SCSI Bus Status   =   %02X",scsi_stat);
				printf("\n05  Bus & Status              =   %02X",status);
				printf("\n06  Input Data                =   %02X",inportb(in_data));

				printf("\nCDB =");
				for (i = 0; i < 16; i++)
					printf (" %02X",cdb[i]);
				scsi_error = true;
				pause_flag = true;
				return;
				}
			break;

		case 1:                                                 /* Reselection */
			if ((status & 0x10) && (scsi_stat & 0x06) &&
			   ((cur_data & parms.mid) == parms.mid))	/* Int, SEL/ & I_O/ asserted & my ID */
				{outportb(init_cmd, 0x08);		/* Assert BSY/ */
				while (inportb(scsi_bus) & 0x02);	/* Poll for SEL/ deasserted */
				outportb(init_cmd, 0);  		/* Deassert BSY/ */
				printf("\nReselection complete");	/* (not able to catch BSY/ false - aprox. 700ns) */
				status = inportb(stat_reg);
				printf("\n05  Bus & Status              =   %02X",status);
				}
			else
				{printf(" ERROR !  isr(2)");
				if ((status & 0x30) == 0x30)
					printf("\n Parity Error");
				printf("\n00  Current SCSI Data   	=   %02X",cur_data);
				printf("\n01  Initiator Command         =   %02X",inportb(init_cmd));
				printf("\n02  Mode                      =   %02X",inportb(mod_reg));
				printf("\n03  Target Command            =   %02X",inportb(tgt_cmd));
				printf("\n04  Current SCSI Bus Status   =   %02X",scsi_stat);
				printf("\n05  Bus & Status              =   %02X",status);
				printf("\n06  Input Data                =   %02X",inportb(in_data));

				printf("\nCDB =");
				for (i = 0; i < 16; i++)
					printf (" %02X",cdb[i]);
				scsi_error = true;
				pause_flag = true;
				return;
				}
			break;

		case 2:						/* Request Sense, variable data block size, Data In */
			while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active */
			scsi_stat = inportb(scsi_bus);

			if((scsi_stat & 0x1C) == 0x04) 		/* 8'th byte of data tells how    */
				{if (parms.detail == 'y')
					printf("extended reply...");	/* many more data bytes to follow */
				mode = 0;
				outportb(mod_reg, mode);		/* Reset DMA & other mode bits */
				initiat = initiat & 0xFE;
				outportb(init_cmd, initiat);		/* Reset Assert Data Bus */
				dma_xfer_data (&dmabyte[8], (dmabyte[7] & 0x00ff), 'r');
				}

			else if((status & 0x80) == 0x80 && (scsi_stat & 0x1C) == 0x0C)		/* EOP Int */
				{if (parms.detail == 'y')
					printf("complete");
				}

			else
				{printf(" ERROR !  isr(2)");
				if ((status & 0x30) == 0x30)
					printf("\n Parity Error");
				printf("\n00  Current SCSI Data   	=   %02X",cur_data);
				printf("\n01  Initiator Command         =   %02X",inportb(init_cmd));
				printf("\n02  Mode                      =   %02X",inportb(mod_reg));
				printf("\n03  Target Command            =   %02X",inportb(tgt_cmd));
				printf("\n04  Current SCSI Bus Status   =   %02X",scsi_stat);
				printf("\n05  Bus & Status              =   %02X",status);
				printf("\n06  Input Data                =   %02X",inportb(in_data));

				printf("\nCDB =");
				for (i = 0; i < 16; i++)
					printf (" %02X",cdb[i]);
				scsi_error = true;
				pause_flag = true;
				return;
				}
			break;



		default:
			printf("\nInput argument error to isr().  phase = %d",phase);
			scsi_error = true;
		}								/* end of switch */
	}									/* end of isr  */
