/********************************** C80S.C *********************************
Copyright Symbios Logic Inc.

Author:		Sherri Moore Symbios Logic System Applications Engineer
Purpose:		53C80S Demonstration Software

This software was written by Symbios Logic to develop and test new
products.  It is not a general SCSI driver.  Symbios Logic assumes no liability for
its use, misuse or inability to use.

Project files:
		C80S.C 		Main body of code
		TARGET80.C	Target Mode routines
		C80SISR.C	Interrupt Service Routine
		C80STEST.C	Automatic Test Module
		DMA80S.C	Hardware dependant DMA code (8237 compatible)
		DMA80S.H  	DMA controller parameters
		C80S_AT.H	Hardware addresses, global defines
		C80S_EV.H	External variables
		C80SCAM.C	SCAM Menu
		C80SCAM.H

rev 1.0	  7-95	Initial release as C80SDEMO
****************************************************************************/
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <dos.h>
#include <process.h>
#include <ctype.h>
#include <string.h>
#include "c80s_at.h"

/***************************** Global Variables **************************/
struct parms {  unsigned 	mid, tid, atn, msg_byte, blksize, blkcount,
				timeout, parity, detail, DMA, ena_eop,
				tar_mode, scsi_reset, clr_reset;

	     } parms;

struct SCAMparms {
		unsigned char 	target, seldelay, priority, assigned_id[8];
		char 		idstring[32];
		 } SCAMparms;

int 		i, j, temp1, temp2, page, pagemax;
unsigned long 	seg, off, abs_adrs, iocount, byte_count;


unsigned int 	dma_buf[16384], cdb_buf[9], sence_data[7], intstat, reg_set,
		msg, x_count, scsicmd, statbyte, inquiry_data[35];

unsigned char 	far* cdb, far* dmabyte, scsi_error, cmd_error, mode, group_code,
		bus_ID, ID_msg, status, cur_data, scsi_stat, initiat,
		tar_phase, cdb_bytes, rcv_byte, pause_flag, TCR;

char 		cmd,temp,looprw,setup_cmd;

/*********************************** main **********************************

inputs:  none
outputs: stdout CRT
calls:   almost everything

Description:  This module prints a menu on the standard output (called CRT
through out this program) and prompts the user to enter a letter from the
menu corresponding to some test to perform.  All of the tests are performed
in other modules.

***************************************************************************/
void main(void)
{
	init ();					/* Init variables, chips, motherboard DMA, */
	while (1)
	{       do						/* clear unexpected interrupts */
		{	status = inportb(stat_reg);
			if (status & 0x10)
			{
			inportb(rst_int);				/*clear interrupts*/
			printf("\n Clear Interrupt.");
			status = inportb(stat_reg);
			if (status == 0xFF)
			     {printf("\nCHIP NOT RESPONDING\n");
			      return;
			     }
			}
			printf("INTERRUPT ");
		}while (status & 0x10);

		clrscr();
		scsi_error = false;
		printf("\n\n DMA start address = %06lX",abs_adrs); 		/* troubleshooting */
		printf("\n\n    ********************* Symbios 53C80S AT Board Test Menu********************\n");
		printf("\n  b .................... Display buffer memory");
		printf("\n  d .................... Init buffer with decrementing pattern");
		printf("\n  e .................... Execute an SCSI command");
		printf("\n  i .....................Init buffer with incrementing pattern");
		printf("\n  j .....................Fill buffer with alternating 00 FF");
		printf("\n  l .....................Loop on read or write");
		printf("\n  m .....................SCAM Menu");
		printf("\n  r .....................Read 53C80S Registers");
		printf("\n  R .....................Reset SCSI bus");
		printf("\n  s .....................Setup menu");
		printf("\n  t .....................Target mode");
		printf("\n  q .................... Quit; exit this program");
		printf("\n\n ? ");

		cmd=getche();

		switch (cmd)
			{case 'b' :
				dumpbuf();			/* display buffer memory */
				break;
			case 'd' :				/* decrement buffer, start at keyboard input*/
				printf("\n Initialize memory buffer with an decrementing pattern");
				printf("\n Enter starting value in hex ? ");
				scanf("%x%c",&j,&temp);	/* temp is to discard carriage return */
				fillbuff(cmd,j);
				break;
			case 'e' :					/* Execute an SCSI command */
				scsi_menu();
				if(cmd_error)
					printf("\n   Unsupported Command !");
				break;
			case 'i' :				/* increment buffer, start at keyboard input */
				printf("\n Initialize memory buffer with an incrementing pattern");
				printf("\n Enter starting value in hex ? ");
				scanf("%x%c",&j,&temp);	/* temp is to discard carriage return */

				fillbuff(cmd,j);
				break;
			case 'j' :				/* init buffer with alternating 0C 0D hex */
				fillbuff(cmd,0);
				break;
			case 'l' :
				clrscr();
				printf("\n\nLoop Read/Write Test\n\n");
				looprw = 'r';
				load_regs();

				cdb[1]=0x00;
				cdb[2]=0x00;
				cdb[3]=0x00;
				cdb[4]=parms.blkcount;
				cdb[5]=0x00;

				do {
					if (looprw == 'r')
						looprw = 'w';
					else
						looprw = 'r';

					if (looprw == 'r')
						cdb[0]=0x08;	/* SCSI Read Command */
					else
						cdb[0] =0x0a;

					DATA_DRIVER(looprw, 6, byte_count);
					} while(!kbhit() && !scsi_error);
				break;
			case 'm' :					/* SCAM menu */
				SCAM_menu();
				pause_flag = false;
				break;
			case 'q' :
				return;					/* quit program */
			case 'r' :
				read_regs();			/* read 53CF92 registers */
				break;

			case 'R' :					/* SCSI Reset */
				outportb(init_cmd, 0x80);
				outportb(init_cmd, 0x00);			init();
				break;
			case 's' :
				setup();				/* parameter Set-up menu */
				break;
			case 't' :
				tar_phase = 0x00;
				target_mode();
				if (scsi_error) {
					printf("\n\nhit any key to continue");
					getch();
					}
				break;
			default:
				printf("\nInvalid response.  Enter a case sensitive letter from the menu.\n");
				delay(1500);
			}					/* end of switch function  */

		if(scsi_error)
			{outportb(init_cmd, 0x80);		/* SCSI RST/ */
			outportb(init_cmd, 0x00);
			load_regs();
			}

		if (pause_flag == true) 		/* pause to view results on CRT */
			{printf("\nhit any key to return to main menu");
			temp = getch();
			pause_flag=false;
			}
	}
}							/* end of main */


/******************************* Initialize *********************************
inputs:  none
outputs: none
calls:   initIBMdma(), load_regs(), init86C01() *microchannel version only*
*************************************************************************/
void init(void) {
	dmabyte = (unsigned char *) dma_buf;	/* align char array to word boudary */
	cdb     = (unsigned char *) cdb_buf;
	parms.mid = 0x02;				/* my ID, SCSI ID 7 */
	parms.tid = 0x01;				/* destination target device ID 0 */
	parms.atn = 'y';				/* Select without ATN */
	parms.msg_byte =0x80;				/* ID Message Byte = 0 */
	parms.blksize = 512;				/* number of bytes in data block */
	parms.blkcount = 1; 				/* number of blocks to transfer */
	parms.detail = 'y';				/* Display lots of info during SCSI */
	parms.parity = 'y';				/* enable automatic parity checking  & int*/
	parms.DMA = 'y';				/* use DMA for data phase */
	parms.ena_eop = 'y';				/* Enable EOP Interrupt*/
	parms.tar_mode = 'n';				/* Target Mode*/
	parms.scsi_reset = 0x80;			/* Set Enable RST/ */
	parms.clr_reset = 0x00;				/* Clear Enable RST/ */
	SCAMparms.target = 1;				/* SCAM Tolerant target device */
	SCAMparms.seldelay = 0x06;			/* selection response delay of 10ms */
	SCAMparms.priority = 1;				/* priority flag for SCAM isolation */
	stpcpy(&SCAMparms.idstring[2],"Symbios ");		/* Vendor ID Code */
	stpcpy(&SCAMparms.idstring[10],"Slave Device 1");	/* Vendor Unique Code */

	outportb(init_cmd, 0x80);
	outportb(init_cmd, 0);		/* Reset */

	load_regs();

	initIBMdma(dmabyte, byte_count, 'w');
	}

/**************** Load Registers in 53C80S SCSI Controller ***************/

void load_regs(void)
	{byte_count = parms.blksize * parms.blkcount;	/* number bytes to send */
	initiat = 0;
	outportb(init_cmd, initiat);

	mode = 0;
	outportb(mod_reg, mode);

	tar_phase = 0;
	outportb(tgt_cmd, tar_phase);
	}


/******************************** fill buffer *******************************
fillbuff(pattern, start)

input arguemnt:	start   = starting value of pattern
		pattern = type of pattern to fill the memory buffer with.

			  i = incrementing pattern starting at keyboard input.
			  d = decrementing pattern starting at keyboard input.
			  j = alternating 00 FF (hex).

**************************************************************************/
void fillbuff(unsigned char pattern, unsigned int start) {
	byte_count = parms.blksize * parms.blkcount;
	switch (pattern) {
		case 'i' :
			for (i=0; i<byte_count; ++i)
				dmabyte[i] = start + i;
			break;
		case 'd' :
			for (i=0; i<byte_count; ++i)
				dmabyte[i] = start - i;
			break;
		case 'j' :
			for (i=0; i<byte_count; i=i+2)
				dmabyte[i] = 0x00;
			for (i=1; i<byte_count; i=i+2)
				dmabyte[i] = 0xFF;
		}								/* end of switch statement */
	}									/* end of fillbuf() */

/******************************* dump buffer ********************************
dumpbuf()

This routine prints the memory array out to the CRT.  There are a varible
number of pages, each of which has 16 rows of 16 bytes displayed in hex.

				***************/
void dumpbuf(void) {
	pagemax = byte_count >> 8;			/* total-bytes divided by bytes-per-page */
	page = 0;
	clrscr();
	do {
		printf("\n\n");
		for (j = 0; j < 16; j++) {
			printf("\n");
			for (i = 0; i < 16; i++)
				printf("%02X  ",dmabyte[i + (j<<4) + (page<<8)]);
			}							/* end of j loop */
		++page;
		printf("\n\n   %d pages left.  ", pagemax - page);
		printf("\n space bar to continue or q to quit");
		temp = tolower(getch());
		} while (temp != 'q' &&  page < pagemax);
	}									/* end of dumpbuf */

/************************** Parameter Set-Up menu **************************/

void setup(void) {
	do {
		clrscr();
		printf("\n\n     **************************** Set-Up Menu ****************************\n");
		printf("\n  a - select with ATN ....................    %c",parms.atn);
		printf("\n  b - block size (bytes per block)........ %4d",parms.blksize);
		printf("\n  B - block count (blocks to transfer).... %4d",parms.blkcount);
		printf("\t  total bytes: %d",parms.blksize * parms.blkcount);
		printf("\n  c - Check Parity .......................    %c",parms.parity);
		printf("\n  D - DMA Transfers during Data Phase.....    %c",parms.DMA);
		printf("\n  e - Enable EOP Interrupt................    %c",parms.ena_eop);
		printf("\n  i - ID message byte ....................   %02Xh",parms.msg_byte);
		printf("\n  m - My ID ..............................   %02Xh",parms.mid);
		printf("\n  t - Target ID ..........................   %02Xh",parms.tid);
		printf("\n  v - Detailed Messages to CRT ...........    %c",parms.detail);
		printf("\n  q - Exit this menu");
		printf("\n\n ? ");

		setup_cmd = getche();
		switch (setup_cmd) {
			case 'a' :					/* set flag for select with or without Attention */
				do {
					printf("\n Select with Attention  (y/n) ?");
					parms.atn = tolower(getche());
					} while (parms.atn != 'y' && parms.atn != 'n');	/* accept only y or n */
				if(parms.atn == 'y') {
					printf("\n ID Message byte in hex ?");	/* need Message Byte */
					scanf("%x%c", &parms.msg_byte, &temp);
					}
				break;
			case 'b' :					/* change Block Size */
				do {
					printf("\n New Block Size in decimal (4096 max) ? ");
					scanf("%d%c",&parms.blksize,&temp);
					} while (parms.blksize > 16384);
				/* no BREAK statement here to next question */
			case 'B' :					/* change Block Count */
				do {
					printf("\n Block Size * Block Count = 16384 bytes maximum");
					printf("\n New Block Count in decimal (%d max) ? ",16384/parms.blksize);
					scanf("%d%c",&parms.blkcount,&temp);
					} while (parms.blkcount * parms.blksize > 16384);
				break;
			case 'c' :
				do {
					printf("\n Enable automatic Parity Checking ? ");
					parms.parity = tolower(getche());
					} while (parms.parity != 'y' && parms.parity != 'n');
				break;
			case 'D' :
				do {
					printf("\n Use DMA during Data Phase ? ");
					parms.DMA = tolower(getche());
					} while (parms.DMA != 'y' && parms.DMA != 'n');
				break;
			case 'e' :
				do {
					printf("\n Enable EOP Interrupt ? ");
					parms.ena_eop = tolower(getche());
					} while (parms.ena_eop != 'y' && parms.ena_eop != 'n');
				break;
			case 'i' :					/* change ID Message Byte */
				printf("\nnew ID Message Byte in hex ? ");
				scanf("%x%c",&parms.msg_byte,&temp);
				break;
			case 'm' : 				/* change My ID */
				do {
					printf("\n New ID of this machine in hex ? ");
					scanf("%02X%c",&parms.mid,&temp);
					} while (parms.mid > 0x80);
				break;
			case 'q' :
				break;				/* need a place holder to avoid default */
			case 't' :					/* change Target ID */
				do 	{
					printf("\n New Target ID in hex ? ");
					scanf("%02X%c",&parms.tid,&temp);		/* temp is to discard CRLF */
					} while (parms.tid > 0x80);
				break;
			case 'v' :
				printf("\n Verbose mode during SCSI (y/n) ? ");
				temp = tolower(getche());
				switch (temp)
					{
					case 'n' :
						parms.detail = temp;
						break;
					case 'y' :
						parms.detail = temp;
						break;
					default:
						printf("\n\nDetail setting not changed");
					} 					/* end of get character case statement */
				break;
			default:
				printf("Invalid response.  Enter a case sensitive letter from the menu.\n");
				delay(1500);
			}							/* end of switch(setup_cmd) */
		} while (setup_cmd != 'q');						/* end of do-while loop */

	load_regs();								/* load the parameters into the chip */
	while(!kbhit());
	temp = getche();
	}									/* end of setup() */



/******************************** SCSI menu ********************************/
void scsi_menu(void) {
	scsi_error = false;
	cmd_error = false;
	scsicmd = 0;

	clrscr();
	printf("\n****************************** SCSI Command Menu *****************************\n");
	printf("\n  00 ............................. Test Unit Ready");
	printf("\n  03 ............................. Request Sense");
	printf("\n  08 ............................. Read");
	printf("\n  0a ............................. Write");
	printf("\n  12 ............................. Inquiry");
	printf("\n  28 ............................. Extended Read");
	printf("\n  2a ............................. Extended Write");
	printf("\n\n ? ");
	scanf("%x",&scsicmd);

	switch(scsicmd) {
		case 0x00 :
			tur();     					/* Test Unit Ready */
			break;
		case 0x03 :					/* Request Sense */
			printf("\nRequest Sense  (display buffer memory to see sense data)\n");
			cdb[0] = 0x03;				/* op-code */
			cdb[1] = 0x00;				/* lun = 0 */
			cdb[2] = 0x00;				/* reserved */
			cdb[3] = 0x00;				/* reserved	*/
			cdb[4] = 0xff;				/* allocation length */
			cdb[5] = 0x00;				/* control byte */

			rqsense(6);				/* driver for Request Sense */
			break;
		case 0x08 :        				/* SCSI read command */
			cdb[0] = 0x08;
			cdb[1] = 0x00;				/* lun = 0, msbaddr = 0	*/
			cdb[2] = 0x00;				/* msb address = 00	*/
			cdb[3] = 0x00;				/* lsb address = 00	*/
			cdb[4] = parms.blkcount;			/* number blocks to transfer */
			cdb[5] = 0x00;				/* reserved byte */

			DATA_DRIVER ('r', 6, byte_count);	/* driver for data transfer*/
			break;
		case 0x01 :                     		/* SCSI read command-delay REQs */
			cdb[0] = 0x08;
			cdb[1] = 0x00;
			cdb[2] = 0x01;
			cdb[3] = 0x00;
			cdb[4] = parms.blkcount;
			cdb[5] = 0x01;              			/* code to add delay between REQs */

			DATA_DRIVER('r',6,byte_count);
			break;

		case 0x0a :					/* SCSI write command */
			cdb[0] = 0x0a;				/* opcode */
			cdb[1] = 0x00;				/* lun = 0, msbaddr = 0	*/
			cdb[2] = 0x00;				/* msb address = 00	*/
			cdb[3] = 0x00;				/* lsb address = 00	*/
			cdb[4] = parms.blkcount;			/* number blocks to transfer */
			cdb[5] = 0x00;				/* reserved byte */

			DATA_DRIVER ('w', 6, byte_count);	/* driver for data transfer*/
			break;
		case 0x12 :					/* SCSI inquiry command */
			printf("\nInquiry Command\n");
			inquiry();
			break;
		case 0x28 : 					/* Extended Read */
			cdb[0] = 0x28;				/* Extended Read Command */
			cdb[1] = 0x00;				/* lun = 0 */
			cdb[2] = 0x00;				/* MSB of block address	*/
			cdb[3] = 0x00;				/* 2SB of block address	*/
			cdb[4] = 0x01;				/* 3SB of block address	*/
			cdb[5] = 0x00;				/* LSB of block address	*/
			cdb[6] = 0x00;				/* reserved byte */
			cdb[7] = parms.blkcount>>8;		/* MSB of transfer length */
			cdb[8] = parms.blkcount;		/* LSB of transfer length */
			cdb[9] = 0x00;				/* control byte	*/

			DATA_DRIVER ('r', 10, byte_count);	/* driver for data transfer*/
			break;
		case 0x2a : 					/* Extended Write Bench Mark Test */
			cdb[0] = 0x2a;				/* Exteded Write Command */
			cdb[1] = 0x00;				/* lun = 0 */
			cdb[2] = 0x00;				/* MSB of block address	*/
			cdb[3] = 0x00;				/* 2SB of block address	*/
			cdb[4] = 0x01;				/* 3SB of block address	*/
			cdb[5] = 0x00;				/* LSB of block address	*/
			cdb[6] = 0x00;				/* reserved byte */
			cdb[7] = parms.blkcount>>8;		/* MSB of transfer length */
			cdb[8] = parms.blkcount;		/* LSB of transfer length */
			cdb[9] = 0x00;				/* control byte	*/

			DATA_DRIVER ('w', 10, byte_count); 	/* driver for data transfer*/
			break;
		default :
			cmd_error = true;				/* unsupported command */
		}						/* end of switch statement */
	}							/* end of scsimenu */


/******************************* Test Unit Ready ************************/

void tur (void) {
	pause_flag = true;
	printf("\nTest Unit Ready Command\n");
	cdb[0] = 0x00;				/* op-code */
	cdb[1] = 0x00;				/* lun = 0 */
	cdb[2] = 0x00;				/* reserved	*/
	cdb[3] = 0x00;				/* reserved	*/
	cdb[4] = 0x00;				/* reserved	*/
	cdb[5] = 0x00;				/* control byte	*/

	select(6);				/* select a device */
	if (scsi_error)
		return;

	cmdcomp();				/* Get Status, Msg, accept & disconnect */
	if (scsi_error)
		return;				/* exit, abort */


	while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
		{scsi_stat = inportb(scsi_bus);
		status = inportb(stat_reg);
		cur_data  = inportb(scsi_data);
		}
	tar_phase = 0;
	outportb(tgt_cmd, tar_phase);
	}


/******************************* Inquiry Command ************************/
unsigned char inquiry(void)
	{int i;
	pause_flag = true;

	cdb[0] = 0x12;					/* op-code */
	cdb[1] = 0x00;					/* lun = 0 */
	cdb[2] = 0x00;					/* page code */
	cdb[3] = 0x00;					/* reserved	*/
	cdb[4] = 0x24;					/* allocation length = 36 bytes */
	cdb[5] = 0x00;					/* control byte	*/

	select(6);
	if (scsi_error)
		return;

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active (valid Phase) */
	scsi_stat = inportb(scsi_bus);

	while ((scsi_stat & 0x1C) == 0x1C)		/* Message In ? */
		{Message_In();				/* returns w/ valid phase */
		if (scsi_error)
			return(0);
		}

	if ((scsi_stat & 0x1C) != 0x04)         /* Data In Phase? */
		{printf("\n Illegal Phase, expected Msg or Data In");
		printf("\n Current SCSI Bus Status = %02X", inportb(scsi_bus));
		scsi_error = true;
		return(0);
		}			  	/* Illegal Phase, abort */

	if (parms.detail == 'y')
		printf("\n Inquiry Data In ...");
	outportb(tgt_cmd, 1);			/* Set PC bits to Data In Phase */

	dma_xfer_data(dmabyte, 36, 'r');	/* DMA data transfer */

	if (scsi_error)
		return(0);

	cmdcomp();			/* Get Status, Msg, accept & disconnect */
	if (scsi_error)
		return(0);		/* exit, abort */

	while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
		{scsi_stat = inportb(scsi_bus);
		status = inportb(stat_reg);
		cur_data  = inportb(scsi_data);
		}

	if (parms.detail == 'y') {
		printf("\nID: ");
		for (i = 8; i < (dmabyte[4] + 5); i++)
			putch(dmabyte[i]);		/* listing the id string */
		}

	return(0);
	}

/********************************* DATA_DRIVER ******************************
DATA_DRIVER (rw, cdb_len, data_len)

input argument rw = phase target is expected to request.r for Data In phase,
							w for Data Out.

	  cdb_len = number of bytes in Command Descripter Block
	 data_len = number of bytes in data block

This is the driver that calls the four routines that programs the 53C80S to:
Select a device, Transfer data across the SCSI bus, receive Status and
Message bytes from the Target, and disconnect from the SCSI bus.

******************************************************************************/
void DATA_DRIVER(char rw, unsigned short cdb_len, int data_len) {
	pause_flag=true;

	select(cdb_len);				/* select a device */
	if (scsi_error)
		return;

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active (valid Phase) */

	while ((inportb(scsi_bus) & 0x1C) == 0x1C) 		/* Message In ? */
			{
			Message_In();		/* returns w/ valid phase */
			if (scsi_error)
				return;
			}

	if(parms.detail == 'y')
		{if (parms.DMA == 'y')					/* DMA data transfer */
			printf("\n DMA ");
		else
			printf("\n nonDMA ");
		}

	scsi_stat = inportb(scsi_bus);
	if (rw == 'w')
		{if ((scsi_stat & 0x1C) != 0)           /* Data Out Phase? */
			{scsi_error = true;
			return;                         /* Illegal Phase, abort */
			}
		outportb(tgt_cmd, 0);			/* Set PC bits to Data Out */
		if(parms.detail == 'y')
			printf("Data Out ... ");
		}


	else
		{if ((scsi_stat & 0x1C) != 0x04)        /* Data In Phase? */
			{scsi_error = true;
			return;                         /* Illegal Phase, abort */
			}
		outportb(tgt_cmd, 1);			/* Set PC bits to Data In */
		if(parms.detail == 'y')
			printf("Data In ... ");
		}

	if (parms.DMA == 'y')					/* DMA data transfer */
		dma_xfer_data(&dmabyte[0], data_len, rw);

	else 							/* non-DMA data trasfer */
		nonDMA_xfer_data(dmabyte, data_len, rw);

	if (scsi_error)
		return;

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active (valid Phase) */

	while ((inportb(scsi_bus) & 0x1C) == 0x1C) 		/* Message In ? */
			{
			Message_In();		/* returns w/ valid phase */
			if (scsi_error)
				return;
			}

	cmdcomp();		/* Get Status, Msg, accept & disconnect */
	if (scsi_error)
		return;		/* exit, abort */

	while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
		{scsi_stat = inportb(scsi_bus);
		status = inportb(stat_reg);
		cur_data  = inportb(scsi_data);
		}
	}


/********************************** select *********************************/
/* Arbitrates for the bus until wins, Selects, Sends Identify Message Byte if
Selected with ATN/ & Sends CDB */

void select(unsigned short cdb_len) 		  /* Select an SCSI device */
	{
	scsi_error = false;

	outportb(tgt_cmd, 0);
	outportb(sel_ena, parms.mid);		/* Enable Selection-Reselection */
	outportb(scsi_data, parms.mid);		/* ID bit to Output Data Reg */


/* test         outportb(init_cmd, 0x08);	/* Set BSY/ to test Arb */


arb:	mode = 0x01;
	outportb(mod_reg, mode);                /* Begin Arbitration */


/* test		while (!kbhit());
		getch();
		outportb(init_cmd, 0);	  /* Clear BSY/ to start arb */



	while (!(inportb(init_cmd) & 0x40));	/* While Arb in Process = 0 */

	/* Wait 2.2 usec (min) Arbitration Delay here */

	if (inportb(init_cmd) & 0x20)		/* IF Lost Arbitration */
		{mode = 0x00;
		outportb(mod_reg, mode);	/* Reset the Arb bit */
		goto arb;
		}

	if ((parms.mid ^ inportb(scsi_data)) > parms.mid)  /* If Higher Id on Bus lose arb (^ bitwise exclusive or to get rid of my ID */
		{mode = 0x00;
		outportb(mod_reg, mode);	/* Reset the Arb bit */
		goto arb;
		}

	initiat = 0x04;
	outportb(init_cmd, initiat);		/* Assert SEL/ (initiat was 0) */
	if(parms.atn == 'y')
		{initiat = initiat | 0x02;
		outportb(init_cmd, initiat);	/* Assert ATN/ */
		}

	if (inportb(init_cmd) & 0x20)		/* IF Lost Arbitration */
		{initiat = 0;
		outportb(init_cmd, initiat);	/* Reset SEL/ & ATN/ */
		mode = 0x00;
		outportb(mod_reg, mode);	/* Reset the Arb bit */
		goto arb;
		}

	/* Wait 1.2 usec (min) Bus Clear & Settle Delay here */

	bus_ID = parms.mid | parms.tid;
	outportb(scsi_data, bus_ID);		/* Target & Init IDs on bus */
	initiat = initiat | 0x09;
	outportb(init_cmd, initiat);		/* Assert BSY/ & Data Bus (do not change SEL/ & ATN/ */
	mode = mode & 0xFE;
	outportb(mod_reg, mode);		/* Reset Arb bit */
	initiat = initiat & 0xF7;
	outportb(init_cmd, initiat);		/* Reset Assert BSY/ */

	if (parms.detail == 'y')
		printf("\n\n\n Waiting for Selection Interrupt... ");
	isr(0);
	if (scsi_error)
		return;

	delay(200);				/* Wait 250ms from BSY deasserted */
						/* ( isr(0) used some of this time */
	if ((inportb(scsi_bus) & 0x40) != 0x40)
		{printf("\n Selection Timeout");
		scsi_error = true;
		return(1);                      /* Selection timeout, abort */
		}

	outportb(sel_ena, 0x00);		/* Reset Select Enable Reg */
	initiat = initiat & 0xFA;
	outportb(init_cmd, initiat); 		/* Deassert SEL/ & Data Bus */

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active (valid Phase) */

	if(parms.atn == 'y')
		{
		if(parms.detail == 'y')
			printf("\nSelection with ATN ");	/* non-DMA select with ATN */
		scsi_stat = inportb(scsi_bus);
		if ((scsi_stat & 0x18) != 0x18)		/* Msg Out Phase? */
			{scsi_error = true;
			return;
			}
		initiat = 0;
		outportb(init_cmd, initiat);		/* Clear ATN/ (this is normally done after ACK is asserted & must be done 90ns before ACK/ is asserted */
		outportb(tgt_cmd, 0x06);      		/* Set PC bits = Msg Out */
		outportb(scsi_data, parms.msg_byte);	/* Data for Send */
		Init_REQ_ACK('w');			/* Send Identify Msg */
		}
	else
		if(parms.detail == 'y')
			printf("\nSelection without ATN ");	/* non-DMA select without ATN */

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active */

	scsi_stat = inportb(scsi_bus);
	if ((scsi_stat & 0x08) != 0x08)		/* Command Phase? */
		{scsi_error = true;
		return;
		}

	outportb(tgt_cmd, 0x02);      		/* Set Phase Control bits = Command Phase */
	if(parms.detail == 'y')
		printf("\n Send CDB... ");
	for (i=0; i < cdb_len; ++i)
		{
		outportb(scsi_data, cdb[i]);		/* CDB byte to Send */
		Init_REQ_ACK('w');				/* Send CDB byte i */
		scsi_stat = inportb(scsi_bus);
		if ((scsi_stat & 0x08) != 0x08)		/* Command Phase? */
			{scsi_error = true;
			return;
			}
		}
	if(parms.detail == 'y')
		printf("\ complete");
	}									/* end of subroutine select */

/*********************** Initiator REQ/ ACK/ handshake *********************/

void Init_REQ_ACK(char rw)
	{
	while (((scsi_stat = inportb(scsi_bus)) & 0x20) != 0x20);	/* Poll for REQ/ active */

	if (rw == 'w')
		{initiat = initiat | 0x01;
		outportb(init_cmd, initiat);		/* Assert Data Bus */
		}
	else
		rcv_byte = inportb(scsi_data);		/* Receive byte */

	initiat = initiat | 0x10;
	outportb(init_cmd, initiat);			/* Assert ACK/ */
	while (((scsi_stat = inportb(scsi_bus)) & 0x20) == 0x20);	/* Poll for REQ/ inactive */
	initiat = initiat & 0xEE;
	outportb(init_cmd, initiat);			/* Deassert ACK & Data Bus */
	}

/*************************** DMA Transfer Data ****************************
dma_xfer_data (start_addrs, count, rw)

Input Arguments:
	start_addrs = starting addres in memory of data block
	count 	    = number of bytes to transfer
	rw 	    = direction of data flow.  r for Data In, w for Data Out

This routine will transfer the specified number of bytes across the SCSI
bus.
***********************************************************************/

void dma_xfer_data(char far* start_addrs, int count, char rw)
	{
	--count;				/* to get EOP w/ last byte */
	initIBMdma(start_addrs, count, rw);	/* address, byte count, direction */

	if (rw == 'w')
		{
		mode = 0x02;		    	/* DMA Mode */
		if (parms.ena_eop == 'y')
			mode = mode | 0x08;	/* Enable EOP Int */

		outportb(mod_reg, mode);

/* Set PC bits prior to calling dma_xfer_data */

		initiat = initiat | 0x01;
		outportb(init_cmd, initiat);	/* Assert Data Bus */

		outportb(dma_send, 0);			/* Start DMA Send */
		while (!(inportb(tgt_cmd) & 0x80));     /* Poll for Last byte sent */
		}
	else
		{
		mode = 0x02;            	/* DMA Mode */
		if (parms.ena_eop == 'y')
			mode = mode | 0x08;	/* Enable EOP Int */
		if (parms.parity == 'y')
			mode = mode | 0x30;	/* Enable Parity chking & Int */

		outportb(mod_reg, mode);

/* Set PC bits prior to calling dma_xfer_data */

		outportb(dma_init_rcv, 0);	/* Start DMA Receive */
		}

	isr(2);					/* EOP or End of DMA Interrupt */

	if (scsi_error)
		return;

	mode = 0;
	outportb(mod_reg, mode);		/* Reset DMA & other mode bits */
	initiat = initiat & 0xFE;
	outportb(init_cmd, initiat);		/* Reset Assert Data Bus */
	}


/************************* non-DMA Transfer Data *************************/

void nonDMA_xfer_data(char far* start_addrs, int count, char rw)
	{
	unsigned int out_count;

	out_count = 0;

	while (out_count < count)
		{scsi_stat = inportb(scsi_bus);
		if (rw == 'w')
			{if ((scsi_stat & 0x1C) != 0)
				{scsi_error = true;
				return;
				}

/* Set PC bits prior to calling nonDMA_xfer_data */

			outportb(scsi_data, start_addrs[out_count]);
			Init_REQ_ACK(rw);
			++out_count;
			}
		else
			{if ((scsi_stat & 0x1C) != 4)
				{scsi_error = true;
				return;
				}

			if (parms.parity == 'y')
				mode = 0x30;		/* Enable Parity chking & Int */

			outportb(mod_reg, mode);

/* Set PC bits prior to calling nonDMA_xfer_data */

			Init_REQ_ACK(rw);
			start_addrs[out_count] = rcv_byte;
			++out_count;
			}

		}	/* End while */

	mode = 0;
	outportb(mod_reg, mode);		/* Reset mode bits */

	initiat = initiat & 0xFE;
	outportb(init_cmd, initiat);		/* Reset Assert Data Bus */
	if(parms.detail == 'y')
		printf("complete");
	}
/******************************** cmdcomp **********************************/

void cmdcomp(void) {				/* Get Status, Msg, accept & disconnect*/
	statbyte = 0xFF;
	msg = 0xFF;
	if(parms.detail == 'y')
		printf("\nStatus and Message ... ");
	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active */

	scsi_stat = inportb(scsi_bus);
	if ((scsi_stat & 0x1C) != 0x0C)		/* expect Status Phase */
		{printf("\n ERROR! Current SCSI Bus Status = %02X (expected 4C)", scsi_stat);
		scsi_error = true;
		return;                         /* abort, Illegal Phase */
		}
	outportb(tgt_cmd, 0x03);		/* Set PC bits to Status Phase */
	Init_REQ_ACK('r');
	statbyte = rcv_byte;

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active */

	scsi_stat = inportb(scsi_bus);
	if ((scsi_stat & 0x1C) != 0x1C)		/* expect Msg In Phase */
		{printf("\n ERROR! Current SCSI Bus Status = %02X (expected 5C)", scsi_stat);
		scsi_error = true;
		return;                         /* abort, Illegal Phase */
		}
	outportb(tgt_cmd, 0x07);		/* Set PC bits to Msg In Phase */
	Init_REQ_ACK('r');			/* Get & accept Msg byte */
	msg = rcv_byte;

	if (statbyte == 0 && msg == 0)
		{if(parms.detail == 'y')
			printf("good");
		}
	else if (statbyte == 2 && msg == 0)
		{if(parms.detail == 'y')
			printf("status = 02 CHECK CONDITION");
		}
	else
		{scsi_error = true;
		return;
		}
	}

/********************************* rqsense **********************************

rqsense(cdb_len)

input argurment cdb_len = length of the command descripter block

This routine is a driver for the Request Sense SCSI command

			       ***************/
void rqsense(unsigned short cdb_len) {
	pause_flag = true;
	select(cdb_len);			/* select a device */

	if(scsi_error)
		return;				/* exit, abort */

	temp = parms.ena_eop;			/* save current ena EOP int setting */
	parms.ena_eop = 'y';			/* must depend on EOP int since
						   the phase may not change */

	while (!(inportb(scsi_bus) & 0x20));	/* Poll for REQ/ active */

	scsi_stat = inportb(scsi_bus);
	if ((scsi_stat & 0x1C) != 0x04)         /* Data In Phase? */
		{scsi_error = true;
		return;				/* Illegal Phase, abort */
		}

	if (parms.detail == 'y')
		printf("\n Request Sence Data In ...");
	outportb(tgt_cmd, 1);			/* Set PC bit to Data In Phase */

	dma_xfer_data(&dmabyte[0], 8, 'r');	/* DMA Data In, 8 bytes */

	parms.ena_eop = temp;			/* restore ena EOP int setting to orig */
	if (parms.ena_eop == 'y')
		mode = mode | 0x08;		/* Set to original (Enable EOP Int) */
	outportb(mod_reg, mode);

	if(scsi_error)
		return;				/* exit, abort */

	cmdcomp();				/* Get Status, Msg, accept & disconnect */
	if (scsi_error)
		return;				/* exit, abort */

	while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
		{scsi_stat = inportb(scsi_bus);
		status = inportb(stat_reg);
		cur_data  = inportb(scsi_data);
		}
	}

/*********************** Message In after Command Phase *********************
							DisConnect then ReSelect

Message_In()

input argument:	rw = data direction, r (Read) for Data In, w for Data Out

				 *************/
void Message_In()
	{
	if(parms.detail == 'y')
		printf("\nMessage In ... ");
	outportb(tgt_cmd, 0x07);

	temp1 = inportb(scsi_data);				/* Receive byte */
	initiat = 0x10;
	outportb(init_cmd, initiat);				/* Assert ACK/ */
	while (((scsi_stat = inportb(scsi_bus)) & 0x20) == 0x20);	/* Poll for REQ/ inactive */

	switch (temp1) {			/* temp1 = message-byte */
		case 0x02 :
			if (parms.detail == 'y')
				printf(" ... Save Data Pointers");
			initiat = 0;
			outportb(init_cmd, initiat);	/* Clear ACK for msg accept */
			break;

		case 0x04 :
			if (parms.detail == 'y')
				printf(" ... Disconnect Message");
			initiat = 0;
			outportb(init_cmd, initiat);	/* Clear ACK for msg accept */

			while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
				{scsi_stat = inportb(scsi_bus);
				status = inportb(stat_reg);
				cur_data  = inportb(scsi_data);
				}

			outportb(sel_ena, parms.mid);	/* enable ReSelection */
			printf("\n\n\n Waiting for Reselection Interrupt");
			isr(1);				/* wait for ReSelect */
			if (scsi_error)
				return;

			bus_ID = cur_data;

			while (((scsi_stat = inportb(scsi_bus)) & 0x20) != 0x20);	/* Poll for REQ/ active */

			ID_msg = inportb(scsi_data);					/* Receive byte */
			initiat = 0x10;
			outportb(init_cmd, initiat);					/* Assert ACK/ */
			while (((scsi_stat = inportb(scsi_bus)) & 0x20) == 0x20);	/* Poll for REQ/ inactive */

			if (ID_msg != 0x80)						/* Identify Message Bad */
				{printf("\n unsupported Identify Message");
				scsi_error = true;
				return;
				}

			initiat = 0;
			outportb(init_cmd, initiat);	/* Clear ACK for msg accept */
			break;

		default:
			printf(" ... received unsupported message = %02X hex",temp1);
			scsi_error = true;
			return;

		}	/* end of switch() */
	while (((scsi_stat = inportb(scsi_bus)) & 0x20) != 0x20);	/* Poll for REQ/ active */
									/* Phase valid */
	}	/* end of Message_In() */


/*************************** Read 53C80S Registers **********************

This routine reads the register in the 53CF92

								  ***********/

void read_regs(void)
	{
	clrscr();
	printf("\nThe current 53C80S register contents are: \n");

	printf("\n00  Current SCSI Data                =   %02X",inportb(scsi_data));
	printf("\n01  Initiator Command                =   %02X",inportb(init_cmd));
	printf("\n02  Mode                             =   %02X",inportb(mod_reg));
	printf("\n03  Target Command                   =   %02X",inportb(tgt_cmd));
	printf("\n04  Current SCSI Bus Status          =   %02X",inportb(scsi_bus));
	printf("\n05  Bus & Status                     =   %02X",inportb(stat_reg));
	printf("\n06  Input Data                       =   %02X",inportb(in_data));
#
	printf("\n\nHit any key to return to menu");
	temp = getche();
	} 									/* end of read_regs() */

