
/********************************** C80SCAM.C *******************************
Copyright Symbios Logic Inc.

Author:		Sherri Moore Symbios Logic System Applications Engineer
Purpose:		53C80S Demonstration Software

This software was written by Symbios Logic to develop and test new products.
It is not a general SCSI driver.  Symbios Logic assumes no liability for
its use, misuse or inability to use.

Project files:
		C80S.C 		Main body of code
		TARGET80.C	Target Mode routines
		C80SISR.C	Interrupt Service Routine
		C80STEST.C	Automatic Test Module
		DMA80S.C	Hardware dependant DMA code (8237 compatible)
		DMA80S.H  	DMA controller parameters
		C80S_AT.H	Hardware addresses, global defines
		C80S_EV.H	External variables
		C80SCAM.H



*************************** SCAM Functional Code **************************/

#include<stdio.h>
#include<conio.h>
#include<dos.h>
#include<string.h>
#include"c80s_at.h"
#include"c80s_ev.h"

void SCAM_menu(void) {
	char menu_cmd;
	char 	*vendor_id = &SCAMparms.idstring[2],
		*serial_id = &SCAMparms.idstring[10];

	while (1) {
		clrscr();
		printf("\n\n     ***************************** SCAM Menu ****************************\n");

		printf("\n       t - SCAM Target Mode ...................    ");
		switch (SCAMparms.target) {
			case 0:
				printf("SCAM Level 2");
				break;
			case 1:
				printf("SCAM Tolerant");
				break;
			case 2:
				printf("non-Tolerant");
				break;
			}
		printf("\n       v - Vendor ID Code .....................    %.8s",vendor_id);
		printf("\n       u - Vendor Unique Code .................    %.16s",serial_id);
		printf("\n       q - Exit this menu");
		printf("\n\n ? ");

		menu_cmd = getche();
		switch (menu_cmd)
			{
			case 't' :					/* toggle between target SCAM modes */
				if (++SCAMparms.target == 3)
					SCAMparms.target = 0;
				break;
			case 'v' :					/* vendor id code */
				printf("\nEnter the Vendor ID Code as 8 characters ?");
				scanf("%8s",vendor_id);
				break;
			case 'u' :					/* serial number */
				printf("\nEnter the Vendor Unique Code as up to 21 characters");
				printf("\n ?");
				scanf("%21s",serial_id);
				break;
			case 'q' :
				return;					/* return to main menu */
			default:
				printf("\nInvalid response.  Enter a case sensitive letter from the menu.\n");
				delay(1500);
			}							/* end of switch(menu_cmd) */
		}
	}

/************************ Responds to SCAM Selection *********************/

unsigned char SCAM_selected(void) {
	/* performs a slave's response to a SCAM selection;
		Returns: 0 - SCSI id has been assigned
			 1 - No id assigned yet
			 2 - Config Process Complete has been received */

	unsigned char ac1, ac2,				/* action code quintets */
				  status,
				  stage = 0;		/* current SCAM stage:  0 - Deferred   1 - Sync Received */

	SCAMparms.priority = 1;				/* setting priority bit */

	/************** SCAM Protocol Initiation ********/
	/* tar_phase is = 0x20 */

	outportb(tgt_cmd, 0x24);			/* set MSG */
	outportb(init_cmd, 0x04);			/* set SEL */
	outportb(tgt_cmd, 0x04);			/* Disable SCAM Selection Response bit */
	outportb(tgt_cmd, 0);				/* release MSG */

	while ((inportb(scsi_bus) & 0x10) || (inportb(scsi_bus) & 0x10))
		if (kbhit())					/* check for MSG released */
			return(1);				/*  using wired-or glitch filtering */

	outportb(init_cmd, 0x0C);			/* assert BSY */
	outportb(scsi_data, 0xC0);			/* set bits DB7 & DB6 */
	outportb(tgt_cmd, 0x01);			/* assert I/O */
	outportb(init_cmd, 0xD0);			/* assert data bus */
	outportb(init_cmd, 0x09);			/* release SEL */

	while ((inportb(scsi_bus) & 0x02) || (inportb(scsi_bus) & 0x02));
		if (kbhit())					/* check for SEL released */
			return(1);				/*  using wired-or glitch filtering */

	scsi_stat = inportb(scsi_bus);			/* get bus status */
/*	outportb(scsi_data, 0x80);			/* release DB6 */

	if (!(scsi_stat & 0x08)) 			/* check if no SCAM masters are present */
		{					/* (C_D/ deasserted) */
		if (parms.detail == 'y')
			printf("\nNo SCAM Masters detected...");
		return(1);
		}
	else
		printf("\n SCAM Master present...");

	outportb(scsi_data, 0x80);			/* release DB6 */

	while ((inportb(scsi_data) & 0x40) || (inportb(scsi_data) & 0x40))
		if (kbhit())					/* check for DB6 released */
			return(1);				/*  using wired-or glitch filtering */
	outportb(init_cmd, 0x0D);			/* assert SEL again */

	/*************** End of SCAM Protocol Initiation ***************/

	while (1) {
		if (!stage)				/* if sync pattern has not been received */
			while (SCAM_trans(0) != 0x1F)	/* wait for sync pattern */
				if (kbhit())
					return(1);

		switch (SCAM_trans(0)) {		/* function code */
			case 0x01:					/* set priority flag & isolate */
				SCAMparms.priority = 1;
				if (parms.detail == 'y')
					printf("\nPriority bit set");
			case 0x00:					/* isolate (assign id) */
				switch (SCAM_isolation()) {
					case 0:				/* terminated */
						ac1 = SCAM_trans(0);
						if (ac1 == 0x1F) {	/* if sync pattern */
							stage = 1;
							break;
							}
						ac2 = SCAM_trans(0);
						if (ac2 == 0x1F) {	/* if sync pattern */
							stage = 1;
							break;
							}
						if (ac1 == 0x18) {	/* assign id */
							ac2 = ac2 & 0x07;
							parms.mid = 1 << ac2;
							if (parms.detail == 'y')
								printf("\nAssigned ID %02Xh",parms.mid);
							return(0);
							}
						if (ac1 == 0x14)
							switch (ac2) {
								case 0x18:	/* clear priority bit */
									SCAMparms.priority = 0;
									if (parms.detail == 'y')
										printf("\nPriority bit cleared");
									break;
								case 0x12:	/* locate on */
									if (parms.detail == 'y')
										printf("\nLocate On action code received");
									break;
								case 0x0B:	/* locate off */
									if (parms.detail == 'y')
										printf("\nLocate Off action code received");
									break;
								}
					case 1: case 3:		/* deferred or error */
						stage = 0;
						break;
					case 2:			/* received new sync pattern */
						stage = 1;
						break;
					}
				break;
			case 0x03:				/* config process complete */
				outportb(sel_ena, 0x00);	/* disable Selection-Reselection */
				return(2);
			case 0x1F:				/* received new sync pattern */
				stage = 1;
				break;
			case 0x0F:				/* dominant master contention */
			default:				/* reserved function code */
				stage = 0;
				break;
			}
		}
	}


unsigned char SCAM_trans(unsigned char data) {
/* performs a single SCAM transfer cycle where data is placed on the DB0-4
by the 53C80S and the resulting data bits on the SCSI bus are returned to
the variable output */

	unsigned char output;
	/* Assert Data Bus bit still set */
	outportb(scsi_data, 0xA0 | data);	/* assert data bits, DB5 & 7 */
	outportb(scsi_data, 0x20 | data);	/* release DB7 */

	while ((inportb(scsi_data) & 0x80) || (inportb(scsi_data) & 0x80))
		if (kbhit())					/* wait for DB7 released */
			return(0xFF);				/*  using wired-or glitch filtering */

	output = inportb(scsi_data) & 0x1F;	/* latch data quintet */
	outportb(scsi_data, 0x60 | data);	/* assert DB6 (DB5 still asserted) */
	outportb(scsi_data, 0x40 | data);		/* release DB5 */
	while ((inportb(scsi_data) & 0x20) || (inportb(scsi_data) & 0x20))
		if (kbhit())				/* wait for DB5 released */
			return(0xFF);			/*  using wired-or glitch filtering */

	outportb(scsi_data, 0xC0);			/* assert DB7 (DB6 still asserted)
							& release data */
	outportb(scsi_data, 0x80);			/* release DB6 */
	while ((inportb(scsi_data) & 0x40) || (inportb(scsi_data) & 0x40))
		if (kbhit())					/* wait for DB6 released */
			return(0xFF);				/*  using wired-or glitch filtering */

	return(output);
	}

unsigned char SCAM_isolation(void){
/* executes the isolation stage for a slave using the SCAM_tran function;
	Returns:		0 - Terminate
				1 - Defer
				2 - Received new sync pattern
				3 - Error
	*/
	int i,j;					/* index variables */
	int strsize;					/* id string length */
	unsigned char data;

	/* setting up type code */
	SCAMparms.idstring[0] = 0x23 | (SCAMparms.priority << 7);
	SCAMparms.idstring[1] = parms.mid;			/* soft id */
	strsize = strlen(&SCAMparms.idstring[10]) + 10;

	for (i = 0; i < strsize; i++) {		/* number of bytes in identification string */
		for (j = 7; j >= 0; j--) 	/* start with most-significant bit */
			{
			data = SCAM_trans(1 << ((SCAMparms.idstring[i] >> j) & 0x01));
			if (data == 0x1F)			/* new sync pattern */
				return(2);				/*  received new sync pattern */
			if (data & 0x0C)			/* if DB2 or DB3 are set */
				return(3);			/*  error */
			if (data & 0x10)			/* if DB4 is set */
				return(0);				/*  terminate */
			if (!((SCAMparms.idstring[i] >> j) & 0x01))	/* if a bit is a 0 */
				if (data & 0x02)			/* if DB1 is set */
					return(1);		/*  defer */
			/* else continue */
			}
		}
	data = SCAM_trans(0);			/* end of id string */
	if (data == 0x1F)			/* new sync pattern */
		return(2);			/*  received new sync pattern */
	if (data & 0x0C)			/* if DB2 or DB3 are set */
		return(3);			/*  error */
	if (data & 0x03)			/* if DB0 or DB1 are set */
		return(1);			/*  defer */

	return(0);				/* else terminate */
	}
