/********************************** TARGET80.C *********************************
Copyright Symbios Logic Inc.

Author:		Sherri Moore Symbios Logic System Applications Engineer
Purpose:		53C80S Demonstration Software

This software was written by Symbios Logic to develop and test new products.
It is not a general SCSI driver.  Symbios Logic assumes no liability for
its use, misuse or inability to use.

Project files:
		C80S.C 		Main body of code
		TARGET80.C	Target Mode routines
		C80SISR.C	Interrupt Service Routine
		C80STEST.C	Automatic Test Module
		DMA80S.C	Hardware dependant DMA code (8237 compatible)
		DMA80S.H  	DMA controller parameters
		C80S_AT.H	Hardware addresses, global defines
		C80S_EV.H	External variables
		C80SCAM.H

rev 1.0	  7-95	Initial release as C80SDEMO

***************************** Target Mode Code ****************************/

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include "c80s_at.h"
#include "c80s_ev.h"


void target_mode(void)
{
unsigned char SCAM_status = 0;		/* 0 - id assigned; normal operation
					   1 - no id assigned; may respond to default id
					   2 - no id assigned; disable until SCAM selection
					   or SCSI reset */

clrscr();
printf("\n\n     Target Mode\nhit any key to Exit\n\n");

if (!SCAMparms.target) {		/* SCAM Level 1&2: need ID assignment */
	tar_phase = 0x20;
	outportb(tgt_cmd, tar_phase);   /* Enable SCAM Selection Detection */
	SCAM_status = 1;
	}

do 	{mode = 0x40;				/* Set Target Mode */
	outportb(mod_reg, mode);		/* Set Mode bits */

	if (SCAM_status != 2)
		outportb(sel_ena, parms.mid);		/* Enable Selection-Reselection */
	if (parms.detail == 'y')
		printf("\n\n\n Waiting for Selection Interrupt");
	while (!kbhit() && ((inportb(stat_reg) & 0x10) == 0));	/* Poll for Selection Interrupt */
	TCR = inportb(tgt_cmd);
	scsi_stat = inportb(scsi_bus);
	status = inportb(stat_reg);
	cur_data  = inportb(scsi_data);

	outportb(sel_ena, 0x00);		/* Clear Select Enable */
	inportb(rst_int);			/* Clear Select Int */

/*	if (status == 0x20)
		{printf("\n Parity Error during Selection");
		load_regs();				/* Clear Mode Reg */
/*		scsi_error = true;
		return;
		}

	if (kbhit())
		{temp = getch();			/* flush keyboard char */
/*		return;
		}*/

	if (status == 0x18 && (scsi_stat == 0 || scsi_stat == 0x80) && cur_data == 0)	/* SCSI RST occured (Bus Free) */
		{printf("\n SCSI Reset Detected");
		while (inportb(scsi_bus) & 0x80);	/* Wait for Long RST to clear */
		load_regs();
		if (!SCAMparms.target) 			/* SCAM Level 1&2: need ID assignment */
			{
			tar_phase = 0x20;
			outportb(tgt_cmd, tar_phase);   /* Enable SCAM Selection Detection */
			SCAM_status = 1;
			}
		}
	else if ((TCR & 0x40) == 0x40)	/* SCAM Selection */
		{
		if (parms.detail == 'y')
			printf("\n\nSCAM Selection...");
		SCAM_status = SCAM_selected();
		if (!SCAM_status)
			{
			printf("\n complete (SCAM Selection Disabled)");
			outportb(init_cmd, 0);		/* Release all SCSI signals */
			tar_phase = 0x00;
			outportb(tgt_cmd, tar_phase);   /* Disable SCAM Selection Detection */
			}
		else
			{
			printf("SCAM Selection Enabled");
			tar_phase = 0x20;
			outportb(tgt_cmd, tar_phase);   /* Enable SCAM Selection Detection */
			}
		}
	else if ((status & 0x10) && ((scsi_stat & 0x42) == 0x02) && 		/* Selection or Reselection */
		((cur_data & parms.mid) == parms.mid))	/* Int, BSY/ deasserted, SEL/ asserted & my ID */
		{
		bus_ID = inportb(scsi_data);
		initiat = 0x08;
		outportb(init_cmd, initiat); 		/* Assert BSY/(within 250ms of BSY/ deasserted)*/

		if (parms.detail == 'y')
			{if (status & 0x02)
				printf("\n\nSelected with ATN ");
			else
				printf("\n\nSelected");
			}

		if (SCAM_status) {
			tar_phase = 0x00;
			outportb(tgt_cmd, tar_phase);   /* Disable SCAM Selection Detection */
			SCAM_status = 0;
			if (parms.detail == 'y')
				printf("\n  Default ID confirmed");
			}

		status = inportb(stat_reg);
		if (status & 0x10)			/* If Interrupt check for parity error */
			{
			if (status & 0x30)
				printf("\n Parity Error during bus_ID receive");
			inportb(rst_int);
			}

		if(parms.parity == 'y')
			mode = mode | 0x30;		/* Enable Parity Checking & Int */
		outportb(mod_reg, mode);		/* Set Mode bits */

		if (status & 0x02)			/* ATN/ asserted */
			{tar_phase = 0x06;
			outportb(tgt_cmd, tar_phase);	/* Set Message Out Phase */
			REQ_ACK();
			ID_msg = rcv_byte;		/* get ID Message */
			}

		status = inportb(stat_reg);
		if (status & 0x10)			/* If Interrupt check for parity error */
			{
			if (status & 0x30)
				printf("\n Parity Error during ID_msg receive");
			inportb(rst_int);
			}

		if (status & 0x02)		/* ATN/ asserted */
			{printf("\n ERROR - Extended Messages are not supported");
			mode = mode & 0xBF;
			outportb(mod_reg, mode);	/* Clear Target Mode (Reg. 3, bit 6)*/
			scsi_error = true;
			return;
			}
		else
			{if (parms.detail == 'y')
				printf("\n Command Phase");	/* ATN/ deasserted */
			tar_phase = 0x02;
			outportb(tgt_cmd, tar_phase);	/* Set Command Phase */
			REQ_ACK();
			cdb[0] = rcv_byte;
			group_code = (cdb[0] & 0xe0) >> 5;

			if (group_code == 0)		/* Decode 1st CDB byte for no. of cmd. bytes */
				cdb_bytes = 6;
			else if (group_code < 3)
				cdb_bytes = 10;
			else
				{printf("\n Group Codes 3 thru 7 are not supported");
				scsi_error = true;
				return;
				}
			for (i=1; i<cdb_bytes; ++i)		/* get remaining CDB bytes*/
				{REQ_ACK();
				cdb[i] = rcv_byte;
				}

			status = inportb(stat_reg);
			if (status & 0x30)
				printf("\n Parity Error during cdb receive");
			inportb(rst_int);

			switch (cdb[0])	{      			/* 1st byte = op-code */
				case 0x00:
					Ttest_ready();  	/* Test Unit Ready */
					break;
				case 0x03:
					Trqsense();     	/* Request Sense */
					break;
				case 0x08:
					if (cdb[4] == 0)
						parms.blkcount = 256;
					else
						parms.blkcount = cdb[4];
					Tread();        			/* Data In */
					break;
				case 0x0a:
					if (cdb[4] == 0)
						parms.blkcount = 256;
					else
						parms.blkcount = cdb[4];
					Twrite();	  		/* Data Out */
					break;
				case 0x12:
					Tinquiry();
					break;
				case 0x28:
					parms.blkcount = (cdb[7] << 8) | cdb[8];
					Tread();        			/* Data In */
					break;
				case 0x2a:
					parms.blkcount = (cdb[7] << 8) | cdb[8];
					Twrite();	  		/* Data Out */
					break;
				default:
					printf("\n unsupported command: ");
					for (i = 0; i < cdb_bytes; ++i)
						printf("%02X ",cdb[i]);
					scsi_error = true;
				}					/* end of switch (cdb[0]) */
			if (scsi_error)
				return;

			mode = mode & 0xCF;
			outportb(mod_reg, mode);		/* Disable Parity Checking during send phases */

			if (parms.detail == 'y')
				printf("\nStatus and Message ... ");

			tar_phase = 0x03;
			outportb(tgt_cmd, tar_phase);		/* Set Status Phase */
			outportb(scsi_data, statbyte);    	/* Status byte to Output Data Reg */
			initiat = initiat | 0x01;
			outportb(init_cmd, initiat); 		/* Set Assert Data Bus */
			REQ_ACK();				/* Send Status Byte */

			tar_phase = 0x07;
			outportb(tgt_cmd, tar_phase);		/* Set Message In Phase */
			outportb(scsi_data, 0x00);      	/* Command Complete Message*/
			REQ_ACK();				/* Send Message Byte */

			if (parms.detail == 'y')
				printf("complete");

			tar_phase = tar_phase & 0xF0;
			outportb(tgt_cmd, tar_phase);		/* Deassert REQ/ & Phase lines */
			while (inportb(stat_reg) & 0x01);	/* Wait for ACK/ to deassert */
			initiat = 0;
			outportb(init_cmd, initiat);			/* Deassert BSY/ & all other signals*/

			while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
				{scsi_stat = inportb(scsi_bus);
				status = inportb(stat_reg);
				cur_data  = inportb(scsi_data);
				}

			++ iocount;
			if (parms.detail == 'y')
				{printf("\n Disconnected");
				printf("\n%cnumber IO's = %8lX h    %12lu d",13,iocount,iocount);
				}
			else
				printf("%cnumber IO's = %8lX h	%12lu d",13,iocount,iocount);
			}						/* end of command */
		}
	else
		{printf("\n#2 - Interrupt detected, but not Selection or Reselection");
		if ((status & 0x30) == 0x30)
			printf("\n Parity Error");
		printf("\n  Bus & Status Register    = %02X",status);
		printf("\n  Current SCSI Bus Reg    = %02X",scsi_stat);
		printf("\n  Current SCSI Data Reg   = %02X",cur_data);
		}

	} while (!kbhit());			/* loop forever */

load_regs();					/* Reset Mode Register */
}               				/* end of target_mode */

/************************Req/ Ack/ Handshake*****************************/

void REQ_ACK(void)
	{tar_phase = tar_phase | 0x08;	     	/* Set Req/ without phase change */
	outportb(tgt_cmd, tar_phase);		/* Assert REQ/ */
	while (!(inportb(stat_reg) & 0x01));	/* Poll for ACK/ asserted */
	if (!(inportb(scsi_bus) & 0x04))	/* If Target Receive...*/
		rcv_byte = inportb(scsi_data);	/* Save received byte */
	tar_phase = tar_phase & 0xF7;		/* Clear Req/ without phase change */
	outportb(tgt_cmd, tar_phase);		/* Deassert REQ/ */
	while ((inportb(stat_reg) & 0x01));	/* Poll for ACK/ deasserted */
	}

/************************ Target Test Unit Ready **************************/

void Ttest_ready(void)
	{if (parms.detail == 'y')
		printf("\nTest Unit Ready");
	statbyte = 0x00;				/* ready code */
	}

/************************** Target Request Sense ***************************/

void Trqsense(void) {

	unsigned int out_count;
	out_count = 0;

	sence_data[0] = 0x70;			/* This reply to Request Sense */
	sence_data[1] = 0x00;
	sence_data[2] = 0x05;
	sence_data[3] = 0x00;
	sence_data[4] = 0x00;
	sence_data[5] = 0x41;
	sence_data[6] = 0xff;
	sence_data[7] = 0x00;

	if (parms.detail == 'y')
		printf("\nRequest Sense Data In ... ");
	tar_phase = 0x01;
	outportb(tgt_cmd, tar_phase);				/* Assert Data In Phase bits */

	do	{
		outportb(scsi_data, sence_data[out_count]);   	/* Write Data to Output Reg */
		initiat = initiat | 0x01;
		outportb(init_cmd, initiat);			/* Assert Data Bus bit */
		REQ_ACK();
		++out_count;
		} while (out_count < 8);

	initiat = initiat & 0xFE;
	outportb(init_cmd, initiat);    	/* Reset Assert Data Bus Bit */

	if(parms.detail == 'y')
		printf("complete");
	statbyte = 0x00;

	}

/******************************** Target Inquiry **************************/

void Tinquiry(void) {

	dmabyte[0] = 0x00;
	dmabyte[1] = 0x00;
	dmabyte[2] = 0x00;
	dmabyte[3] = 0x00;
	dmabyte[4] = 0x1F;
	dmabyte[5] = 0x00;
	dmabyte[6] = 0x00;
	dmabyte[7] = 0x00;

	j=2;
	for (i=8; i < 36; ++i)
		{dmabyte[i] = SCAMparms.idstring[j];
		++j;
		}

	if (parms.detail == 'y')
		printf("\n DMA Data In ... ");

	byte_count = 36;

	mode = mode & 0xCF;		/* Disable Parity Checking during send phases */
	mode = mode | 0x08;		/* Enable EOP Interrupt*/
	outportb(mod_reg, mode);

	--byte_count;				/* To get EOP with last byte */
	initIBMdma(dmabyte, byte_count, 'w');

	mode = mode | 0x02;
	outportb(mod_reg, mode);        /* Assert DMA mode bit */
	tar_phase = 0x01;
	outportb(tgt_cmd, tar_phase);	/* Assert Data In Phase bits */
	initiat = initiat | 0x01;
	outportb(init_cmd, initiat);	/* Assert Data Bus bit */
	outportb(dma_send, 0x00);	/* Write to Start DMA Send */

	if (parms.ena_eop == 'y')
		{while (((status = inportb(stat_reg)) & 0x10) == 0x00);	/* Poll for Interrupt */
		if ((status & 0x80) == 0x80)
			{if (parms.detail == 'y')
				printf(" EOP Interrupt... ");
			inportb(rst_int);
			}
		else
			{printf(" *** Unknown Interrupt *** ");
			if ((status & 0x30) == 0x30)
				printf("\n Parity Error");
			printf("\n  Bus & Status Register    = %02X",status);
			printf("\n  Current SCSI Bus Reg    = %02X",scsi_stat);
			printf("\n  Current SCSI Data Reg   = %02X",cur_data);
			inportb(rst_int);
			return;
			}
		}

	while ((inportb(scsi_bus) & 0x20) && (inportb(stat_reg) & 0x01));   /* Poll for REQ & ACK clear */

	mode = mode & 0xF5;

	outportb(mod_reg, mode);        /* Reset DMA Mode bit */

	initiat = initiat & 0xFE;
	outportb(init_cmd, initiat);    /* Reset Assert Data Bus Bit */

	if (scsi_error) {
		statbyte = 0x02;
		return;
		}

	if(parms.detail == 'y')
		printf("complete");
	statbyte = 0x00;
	}

/************************** Target Send (Data In Phase) *************************/

void Tread(void) {
	unsigned int out_count;

	out_count = 0;

	if (ID_msg & 0x40)
		dis_resel();	       		/* disconnect privlidge? */

	if (scsi_error)
		return;

	if (parms.detail == 'y')
		{if (parms.DMA == 'y')
			printf("\n DMA Data In ...");
		else
			printf("\n Data In ... ");
		}

	byte_count = ((long) parms.blksize) * parms.blkcount;

	mode = mode & 0xCF;
	outportb(mod_reg, mode);		/* Disable Parity Checking during send phases */

	if(parms.ena_eop ==  'y')
		mode = mode | 0x08;		/* Enable EOP Interrupt*/

	if (parms.DMA == 'y')
		{--byte_count;			/* To get EOP with last byte */
		initIBMdma(dmabyte, byte_count, 'w');

		mode = mode | 0x02;
		outportb(mod_reg, mode);        /* Assert DMA mode bit */
		tar_phase = 0x01;
		outportb(tgt_cmd, tar_phase);	/* Assert Data In Phase bits */
		initiat = initiat | 0x01;
		outportb(init_cmd, initiat);	/* Assert Data Bus bit */
		outportb(dma_send, 0x00);	/* Write to Start DMA Send */

		if (parms.ena_eop == 'y')
			{while (((status = inportb(stat_reg)) & 0x10) == 0x00);	/* Poll for Interrupt */
			if ((status & 0x80) == 0x80)
				{if (parms.detail == 'y')
					printf(" EOP Interrupt... ");
				inportb(rst_int);
				}
			else
				{printf(" *** Unknown Interrupt *** ");
				if ((status & 0x30) == 0x30)
					printf("\n Parity Error");
				printf("\n  Bus & Status Register    = %02X",status);
				printf("\n  Current SCSI Bus Reg    = %02X",scsi_stat);
				printf("\n  Current SCSI Data Reg   = %02X",cur_data);
				inportb(rst_int);
				return;
				}
			}
		else
			{while ((inportb(stat_reg) & 0x80) == 0x00);	/* Poll for End of DMA bit */
			if (parms.detail == 'y')
				printf(" End of DMA Transfer... ");
			}

		while ((inportb(scsi_bus) & 0x20) || (inportb(stat_reg) & 0x01));   /* Poll for REQ & ACK clear */

		mode = mode & 0xFD;
		outportb(mod_reg, mode);        		/* Reset DMA Mode bit */

		initiat = initiat & 0xFE;
		outportb(init_cmd, initiat);    /* Reset Assert Data Bus Bit */
		}

	else {  				/****** Non-DMA Transfer *****/
		tar_phase = 0x01;
		outportb(tgt_cmd, tar_phase);	/* Assert Data In Phase bits */

		do	{
			outportb(scsi_data, dmabyte[out_count]);   /* Write Data to Output Reg */
			initiat = initiat | 0x01;
			outportb(init_cmd, initiat);	/* Assert Data Bus bit */
			REQ_ACK();
			++out_count;
			} while (out_count < byte_count);

		initiat = initiat & 0xFE;
		outportb(init_cmd, initiat);    /* Reset Assert Data Bus Bit */
		}

	if (scsi_error) {
		statbyte = 0x02;
		return;
		}

	if(parms.detail == 'y')
		printf("complete");
	statbyte = 0x00;
	}



/****************************** Target Recieve (Data Out Phase) **********************/

void Twrite(void)

	{
	unsigned int out_count;

	out_count = 0;

	if (ID_msg & 0x40)
		dis_resel();	       		/* disconnect privlidge? */

	if (scsi_error)
		return;

	if (parms.detail == 'y')
		{if (parms.DMA == 'y')
			printf("\n DMA Data Out ...");
		else
			printf("\n Data Out ... ");
		}

	byte_count = ((long) parms.blksize) * parms.blkcount;

	mode = mode & 0xCF;
	outportb(mod_reg, mode);		/* Disable Parity Checking during send phases */

	if(parms.ena_eop ==  'y')
		mode = mode | 0x08;		/* Enable EOP Interrupt*/

	if (parms.DMA == 'y')
		{--byte_count;			/* To get EOP with last byte */
		initIBMdma(dmabyte, byte_count, 'r');

		mode = mode | 0x02;
		outportb(mod_reg, mode);        /* Assert DMA mode bit */
		tar_phase = tar_phase & 0xF8;
		outportb(tgt_cmd, tar_phase);	/* Assert Data Out Phase bits */

		outportb(dma_tgt_rcv, 0x00);	/* Write to Start DMA Receive */

		if (parms.ena_eop == 'y')
			{while (((status = inportb(stat_reg)) & 0x10) == 0x00);	/* Poll for Interrupt */
			if ((status & 0x80) == 0x80)
				{if (parms.detail == 'y')
					printf(" EOP Interrupt... ");
				inportb(rst_int);
				}
			else
				{printf(" *** Unknown Interrupt *** ");
				if ((status & 0x30) == 0x30)
					printf("\n Parity Error");
				printf("\n  Bus & Status Register    = %02X",status);
				printf("\n  Current SCSI Bus Reg    = %02X",scsi_stat);
				printf("\n  Current SCSI Data Reg   = %02X",cur_data);
				inportb(rst_int);
				}
			}
		else
			{while ((inportb(stat_reg) & 0x80) == 0x00);	/* Poll for End of DMA bit */
			if (parms.detail == 'y')
				printf(" End of DMA Transfer... ");
			}

		mode = mode & 0xFD;
		outportb(mod_reg, mode);	/* Reset DMA Mode bit */
		}

	else 	{  				/****** Non-DMA Transfer *****/
		tar_phase = tar_phase & 0xF8;
		outportb(tgt_cmd, tar_phase);	/* Assert Data Out Phase bits */

		do 	{
			REQ_ACK();
			dmabyte[out_count] = rcv_byte;   /* Save the received byte */
			++out_count;
			} while (out_count < byte_count);
		}

	if (scsi_error)
		{
		statbyte = 0x02;
		return;
		}

	if(parms.detail == 'y')
		printf("complete");
	statbyte = 0x00;

	}

/******************** Target Disconnect then ReSelect ********************/

void dis_resel(void)
	{if (parms.detail =='y') {
		printf("\nMessage In: Save Data pointers and Disconnect Message ... ");
		}

	mode = mode & 0xCF;
	outportb(mod_reg, mode);		/* Disable Parity Checking during send phases */

	tar_phase = tar_phase | 0x07;
	outportb(tgt_cmd, tar_phase);		/* Set Message In Phase */
	outportb(scsi_data, 0x02);	    	/* Save Data pointers msg to Output Data Reg */
	initiat = initiat | 0x01;
	outportb(init_cmd, initiat); 		/* Set Assert Data Bus */
	REQ_ACK();				/* Send Status Byte */

	outportb(scsi_data, 0x04);      	/* Command Complete Message*/
	REQ_ACK();				/* Send Message Byte */

	if (parms.detail == 'y')
		printf("complete");

	tar_phase = 0;
	outportb(tgt_cmd, tar_phase);		/* Deassert REQ/ & Phase lines */
	while (inportb(stat_reg) & 0x01);	/* Wait for ACK/ to deassert */
	initiat = 0;
	outportb(init_cmd, initiat);			/* Deassert BSY/ & all other signals*/

	while ((status & 0x03) || scsi_stat || cur_data)	/* Poll for Bus Free */
		{scsi_stat = inportb(scsi_bus);
		status = inportb(stat_reg);
		cur_data  = inportb(scsi_data);
		}

	++ iocount;
	if (parms.detail == 'y')
		{printf("\n Disconnected");
		printf("\n%cnumber IO's = %8lX h    %12lu d",13,iocount,iocount);
		}

	if (scsi_error)
		return;

	for (i=800; i>0; ++i) {};		/* Time Delay before Reselect */

	if (parms.detail == 'y')
		printf("\n\nReSelect ... ");

	outportb(sel_ena, parms.mid);		/* Enable Selection-Reselection */
	outportb(scsi_data, parms.mid);
arb:	mode = 0x01;
	outportb(mod_reg, mode);                /* Begin Arbitration */
	while (!(inportb(init_cmd) & 0x40));	/* While Arb in Process = 0 */

	/* Wait 2.2 usec (min) Arbitration Delay here */

	if (inportb(init_cmd) & 0x20)		/* IF Lost Arbitration */
		{mode = 0x00;
		outportb(mod_reg, mode);	/* Reset the Arb bit */
		goto arb;
		}

	if (inportb(scsi_data) > parms.mid)	/* If Higher Id on Bus lose arb */
		{mode = 0x00;
		outportb(mod_reg, mode);	/* Reset the Arb bit */
		goto arb;
		}

	initiat = 0x04;
	outportb(init_cmd, initiat);		/* Assert SEL/ (initiat was 0) */

	if (inportb(init_cmd) & 0x20)		/* IF Lost Arbitration */
		{initiat = 0;
		outportb(init_cmd, initiat);	/* Reset SEL/ */
		mode = 0x00;
		outportb(mod_reg, mode);	/* Reset the Arb bit */
		goto arb;
		}

	/* Wait 1.2 usec (min) Bus Clear & Settle Delay here */

	mode = mode | 0x40;
	outportb(mod_reg, mode);		/* Set Target Mode bit */
	tar_phase = 0x01;
	outportb(tgt_cmd, tar_phase);		/* Assert I_O/ bit */

	outportb(scsi_data, bus_ID);		/* Target & Init IDs on bus */
	initiat = initiat | 0x09;
	outportb(init_cmd, initiat);		/* Assert BSY/ & Data Bus */
	mode = mode & 0xFE;
	outportb(mod_reg, mode);		/* Reset Arb bit */
	initiat = initiat & 0xF7;
	outportb(init_cmd, initiat);		/* Reset Assert BSY/ */

	if (parms.detail == 'y')
		printf("\n\n\n Waiting for Reselection Interrupt");
	while (!kbhit() && ((inportb(stat_reg) & 0x10) == 0));	/* Poll for Reselection Interrupt */
	scsi_stat = inportb(scsi_bus);
	status = inportb(stat_reg);
	cur_data  = inportb(scsi_data);
	inportb(rst_int);				/* Clear Reselection Int */

	outportb(sel_ena, 0x00);			/* Reset Select Enable Reg */

	if (kbhit())
		{temp = getch();			/* flush keyboard char */
		return;
		}

	if (scsi_stat & 0x80)				/* SCSI Reset Detected */
		{printf("\n SCSI Reset Detected before Reselect");
		while (inportb(scsi_bus) & 0x80);
		inportb(rst_int);			/* Reset Interrupt & Parity*/
		load_regs();
		scsi_error = true;
		return;
		}

	else if ((status & 0x10) && (scsi_stat & 0x06) &&	/* Reselection */
	     ((cur_data & parms.mid) == parms.mid))
							/* Int, SEL/ & I_O/ asserted & my ID
							   (not able to catch BSY/ false - aprox. 700ns) */
		{initiat = initiat | 0x08;
		outportb(init_cmd, initiat); 		/* Assert BSY/(within 250ms of BSY/ deasserted) */
		initiat = initiat & 0xFA;		/* & deassert SEL/ & Data Bus */
		outportb(init_cmd, initiat);

		if (parms.detail == 'y')
			printf("\n\nReselected");

		tar_phase = 0x07;
		outportb(tgt_cmd, tar_phase);		/* Assert Msg In Phase */
		ID_msg = 0x80;
		outportb(scsi_data, ID_msg);
		initiat = initiat | 0x01;
		outportb(init_cmd, initiat);		/* Assert Data Bus */
		REQ_ACK();
		}

	else if ((status & 0x03) == 0 && scsi_stat == 0 && cur_data == 0)	/* Bus Free */
		{printf("\n Disconnected (During Reselect) after unknown Interrupt (RST/ ?)");
		if ((status & 0x30) == 0x30)
			printf("\n Parity Error");
		printf("\n  Bus & Status Register   = %02X",status);
		printf("\n  Current SCSI Bus Reg    = %02X",scsi_stat);
		printf("\n  Current SCSI Data Reg   = %02X",cur_data);
		inportb(rst_int);			/* Reset Interrupt & Parity*/
		load_regs();
		scsi_error = true;
		return;
		}
	else
		{printf("\n#2 - Interrupt detected (During Reselect), but not Selection or Reselection");
		if ((status & 0x30) == 0x30)
			printf("\n Parity Error");
		printf("\n  Bus & Status Register    = %02X",status);
		printf("\n  Current SCSI Bus Reg    = %02X",scsi_stat);
		printf("\n  Current SCSI Data Reg   = %02X",cur_data);
		inportb(rst_int);			/* Reset Interrupt & Parity*/
		load_regs();
		scsi_error = true;
		return;
		}

	}

